/* -*- mode: c; c-basic-offset: 4; -*- */
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "daqc.h"
#include "nds2.h"
#include "daqc_response.h"
#include "daqc_internal.h"

#ifdef HAVE_SASL
#include <sasl/sasl.h>
#include <sasl/saslutil.h>

static int
nds2_gets(int fd, char* buf, int buflen) {
    int flags=0;
    int l=0;
    while (l<buflen) {
	int rc = recv(fd, buf+l, buflen-l, flags);
	if (rc == 0) {
	    printf("nds2_gets: Unexpected EOF\n");
	    break;
	} else if (rc < 0) {
	    perror("nds2_gets error");
	    l = -1;
	    break;
	}
	l += rc;
	if (!buf[l-1] || buf[l-1] == '\n') {
	    buf[--l] = 0;
	    break;
	}
    }
    if (l > 0) {
	unsigned int len = 0;
	sasl_decode64(buf, l, buf, buflen, &len);
	l = len;
    } else {
	printf("nds2_gets: Error receiving string: %i\n", l);
    }
    return l;
}

static int
nds2_puts(int fd, const char* buf, int len) {
    int   ltmp = ((len / 3) + 1) * 4 + 2;
    char* tbuf = malloc(ltmp);
    unsigned int ldat = 0;
    sasl_encode64(buf, len, tbuf, ltmp, &ldat);
    tbuf[ldat++] = '\n';
    int rc = send(fd, tbuf, ldat, 0);
    free(tbuf);
    return rc;
}
#endif

/*
 *  User authentication
 */
int
nds2_authenticate(daq_t *daq, const char* server) {
    int result = 0;
#ifdef HAVE_SASL
    /*  Allocate input buffer   */
    int buflen = 2048;
    char* buf  = malloc(buflen);

    /*  Get mechanism list      */
    int nch = nds2_gets(daq->sockfd, buf, buflen);
    if (nch < 0) {
	free(buf);
	return DAQD_ERROR;
    }

    /* client new connection */
    sasl_conn_t* conn = 0;
    result=sasl_client_new("nds2",     /* The service we are using */
			   server,     /* The fully qualified domain
					  name of the server we're
					  connecting to */
			   NULL, NULL, /* Local and remote IP
					  address strings
					  (NULL disables mechanisms
					  which require this info)*/
			   NULL,       /* connection-specific
					  callbacks */
			   0,          /* security flags */
			   &conn       /* alloc on success */
			   );
    /* check to see if that worked */
    if (result != SASL_OK) {
	printf("sasl_client_new failed, rc = %i\n", result);
	free(buf);
	return result;
    }
    daq->auth_ctx = conn;

    /*  Start authentication interchange - decide on a mechanism */
    const char *out, *mechusing;
    unsigned int outlen;
    result=sasl_client_start(daq->auth_ctx,  /* the context from above */ 
			     buf,            /* the list of mechanisms
						from the server */
			     NULL,      /* filled in if an
					   interaction is needed */
			     &out,      /* filled in on success */
			     &outlen,   /* filled in on success */
			     &mechusing
			     );
    if (result != SASL_OK && result != SASL_CONTINUE) {
	printf("error detail: %s\n", sasl_errdetail(daq->auth_ctx));
	free(buf);
	return result;	
    }

    /* Send the mechanism */
    int lmech = strlen(mechusing);
    nds2_puts(daq->sockfd, mechusing, lmech);

    /** interact **/
    while (result == SASL_CONTINUE) {
	/*-------------------  Send the line formatted by SASL */
	nch = nds2_puts(daq->sockfd, out, outlen);
	if (nch < 0) {
	    result = DAQD_ERROR;
	    printf("nds2_authenticate: Error sending client string: %i\n", nch);
	    break;
	}

	/*-------------------  Get a response from the server  */
	nch = nds2_gets(daq->sockfd, buf, buflen);
	if (nch < 0) {
	    result = DAQD_ERROR;
	    printf("nds2_authenticate: Error reading server string: %i\n", nch);
	    break;
	}

	/*-------------------  Turn the crank                   */
	result=sasl_client_step(daq->auth_ctx,   /* our context */
				buf,    /* the data from the server */
				nch,    /* it's length */
				NULL,   /* Client interactions? */
				&out,   /* filled in on success */
				&outlen /* filled in on success */
				);
    }

    /*-----------------------  See how it all went              */
    if (result!=SASL_OK) {
	printf("nds2_authenticate: Error stepping client: %i\n", result);
	free(buf);
	return DAQD_ERROR;
    }

    /*-----------------------  Send the last output line        */
    nds2_puts(daq->sockfd, out, outlen);
    free(buf);
#else
    result = DAQD_ERROR;
#endif
    return result;
}

/*
 *  Connect to the DAQD server on the host identified by `ip' address.
 *  Returns zero if OK or the error code if failed.
 */
int
nds2_connect (daq_t *daq, const char* host, int port) 
{
    /*  Get a socket.
     */
    daq->sockfd = socket (AF_INET, SOCK_STREAM, 0);
    if (daq->sockfd < 0) {
	perror("socket() error");
	return DAQD_SOCKET;
    }

    int on = 1;
    setsockopt (daq->sockfd, SOL_SOCKET, SO_REUSEADDR, 
		(const char *) &on, sizeof (on));

    /*----------------------------------  Get the server address.     */
    int resp = daq_set_server(daq, host, port);
    if (resp) {
	perror("Error in daq_set_server");
	close(daq->sockfd);
	daq->sockfd = -1;
	return resp;
    }

    /*-----------------------------------  Connect to the server      */
    fcntl(daq -> sockfd, F_SETFL, O_NONBLOCK);
    printf("Connecting..");
    fflush(stdout);
#define MAX_CONNECT_RETRY 4
    int connect_rc = 0;
    int cntr;
    for (cntr=0; cntr<MAX_CONNECT_RETRY; ++cntr) {
	connect_rc = connect (daq->sockfd, 
			      (struct sockaddr *) &daq -> srvr_addr, 
			      sizeof (daq -> srvr_addr)
			      );
	if (connect_rc < 0 && (errno == EINPROGRESS || errno == EALREADY)) {
	    sleep(1);
	    printf("..");
	    fflush(stdout);
	} else {
	    break;
	}
    }
    if (connect_rc < 0 && errno != EISCONN) {
	printf(" failed\n");
	perror("connect() error");
	close (daq -> sockfd);
	daq->sockfd = -1;
	return DAQD_CONNECT;
    }

    printf(" authenticate");    
    fcntl(daq -> sockfd, F_SETFL, 0);
    resp = daq_send(daq, "authorize");
    if (resp == DAQD_SASL) {
	printf(" ...");
	resp = nds2_authenticate(daq, host);
    }

    if (resp == DAQD_OK) {
	printf(" done\n");
	daq->datafd = daq->sockfd;
    } else {
	printf(" failed: %s\n", daq_strerror(resp));
	close (daq->sockfd);
	daq->sockfd = -1;
    }
    return resp;
}


/*
 *  Disconnect from the server and close the socket file descriptor
 */
int
nds2_disconnect (daq_t *daq) {
    char *command = "quit;\n";
    int  lcmd = strlen(command);
    if (write (daq -> sockfd, command, lcmd) != lcmd) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf (stderr, "daq_disconnect: write errno=%d\n", errno);
#endif
	return DAQD_WRITE;
    }
    usleep(100000);
#if HAVE_SASL
    sasl_conn_t* conn = (sasl_conn_t*)daq->auth_ctx;
    if (conn) sasl_dispose(&conn);
    daq->auth_ctx = NULL;
#endif
    close (daq -> sockfd);
    return 0;
}

/*
 *  Receive channel data using the NDS2 protocol
 */
int
nds2_recv_channel_list(daq_t *daq, daq_channel_t *channel, int num_channels, 
		   int *num_channels_received, time_t gps, enum chantype type)
{
#define BUF_LENGTH 512
#define MAX_TYPESTR_LENGTH 16
    char buf [BUF_LENGTH];
    char typestr [MAX_TYPESTR_LENGTH];

    /*----------------------------------  Issue the request               */
    if (num_channels != 0) {
	sprintf(buf, "get-channels %i %s;", (int)gps, cvt_chantype_str(type));
    } else {
	sprintf(buf, "count-channels %i %s;",(int)gps, cvt_chantype_str(type));
    }
    int resp = daq_send(daq, buf);
    if (resp) {
	printf("Error return from daq_send(): %s\n", daq_strerror(resp));
	return resp;
    }

    /*----------------------------------  Get the number of channels      */
    int fd = daq->sockfd;
    int channels;
    if (read_uint4(fd, (uint4_type*)&channels) || channels < 0) {
	fprintf (stderr, "Couldn't determine the number of data channels\n");
	return DAQD_ERROR;
    }
    *num_channels_received = channels;
    if (num_channels == 0) return DAQD_OK;

    /*----------------------------------  Loop over channels              */
    int i;
    for (i=0; i < channels; i++) {
	int len=_daq_read_string(fd, BUF_LENGTH, buf);
	if (len <= 0) return DAQD_ERROR;
	if (i < num_channels) {
	    const char* p = buf;
	    char name[MAX_CHANNEL_NAME_LENGTH];
	    _daq_get_string(&p, name, sizeof(name));
	    _daq_get_string(&p, typestr, sizeof(typestr));
	    enum chantype ctype = cvt_str_chantype(typestr);
	    double rate = strtod(p, (char**)&p);
	    _daq_get_string(&p, typestr, sizeof(typestr));
	    daq_data_t dtype = data_type_code(typestr);
	    daq_init_channel(channel + i, name, ctype, rate, dtype);
	}
    }
    return DAQD_OK;
}

/*    Get the name string length
 */
static int
_nds2_name_list_length(daq_t *daq) {
    int i;
    int N = daq->num_chan_request;
    long namel = 0;
    chan_req_t* req = daq->chan_req_list;
    for (i=0; i<N; ++i) {
	namel += strlen(req->name) + 1;
	chantype_t ctype = req->type;
	if (ctype != cUnknown) namel += strlen(cvt_chantype_str(ctype)) + 1;
	req++;
    }
    return namel;
}

/*    Get the name string length
 */
static int
_nds2_get_name_list(daq_t *daq, char* p) {
    int i = 0;
    int j = 0;
    int N = daq->num_chan_request;
    p[i++] = '{';
    chan_req_t* req = daq->chan_req_list;
    for (j=0; j<N; j++) {
	if (j) p[i++] = ' ';
	const char* pchn = req->name;
	while (*pchn) p[i++] = *pchn++;
	chantype_t ctype = req->type;
	if (ctype != cUnknown) {
	    p[i++] = ',';
	    pchn = cvt_chantype_str(ctype);
	    while (*pchn) p[i++] = *pchn++;
	}
	req++;
    }
    p[i++] = '}';
    return i;
}

/*
 *  Receive source list using the NDS2 protocol
 */
int 
nds2_recv_source_list(daq_t *daq, char* list, long max_len, time_t gps, 
		      long* str_len) {

    *str_len = 0;
    int namel = _nds2_name_list_length(daq);

    /*----------------------------------  Build up a command
     */
    char* cmd = "get-source-list";

    /*----------------------------------  Get the text length, allocate a buffer
     */
    int N = daq->num_chan_request;
    long txt_len = strlen(cmd)
	         + strlen(" xxxxxxxxxx {};") 
	         + namel + N;
    char* cmd_txt = (char*)malloc(txt_len);

    /*----------------------------------  Build the command
     */
    char* p = cmd_txt;
    while (*cmd) *p++ = *cmd++;
    *p++ = ' ';
    p += _daq_cvt_string(gps, p);
    *p++ = ' ';
    p += _nds2_get_name_list(daq, p);
    *p++ = ';';
    *p++ = 0;  

    /*----------------------------------  Send the command free the command
     */
    int rc   = daq_send(daq, cmd_txt);
    free(cmd_txt);
    cmd_txt = 0;

    if (rc) {
	printf("Error in daq_send: %s\n", daq_strerror(rc));
	return rc;
    }

    /*----------------------------------  Read the source string      */
    *str_len = _daq_read_string(daq->sockfd, max_len, list);
    if (*str_len < 0) return DAQD_ERROR;
    return rc;
}

/*  nds2_request_data(daq_t*, time_t, time_t)
 *
 *  Get requested channel data for the specified interval.
 */
int
nds2_request_data(daq_t* daq, time_t start, time_t end, time_t dt) {

    /*  Get the request type and sum channel name length
     */
    int i;
    int N = daq->num_chan_request;
    if (N == 0) {
#ifdef DAQC_ACCESS_VERBOSE_ERRORS
	fprintf(stderr, "No channels requested\n");
#endif
	return DAQD_ERROR;
    }

    int namel = _nds2_name_list_length(daq);

    /*----------------------------------  Build up a command
     */
    char* cmd = "get-data";

    /*----------------------------------  Get the text length, allocat a buffer
     */
    long txt_len = strlen(cmd)
	         + strlen(" xxxxxxxxxx xxxxxxxxx xxxxxxxxxx {};") 
	         + namel + N;
    char* cmd_txt = (char*)malloc(txt_len);

    /*----------------------------------  Build the command
     */
    char* p = cmd_txt;
    while (*cmd) *p++ = *cmd++;
    *p++ = ' ';
    p += _daq_cvt_string(start, p);
    *p++ = ' ';
    p += _daq_cvt_string(end, p);
    *p++ = ' ';
    p += _daq_cvt_string(dt, p);
    *p++ = ' ';
    p += _nds2_get_name_list(daq, p);
    *p++ = ';';
    *p++ = 0;  

    /*----------------------------------  Reset status words before request
     */
    chan_req_t* req = daq->chan_req_list;
    for (i=0; i<N; ++i) {
	req[i].status = 0;
    }

    /*----------------------------------  Send the command free the command
     */
    int rc   = daq_send(daq, cmd_txt);
    free(cmd_txt);
    cmd_txt = 0;


    if (rc) {
	printf("Error in daq_send: %s\n", daq_strerror(rc));
	return rc;
    }

    /*------------------------------  Get the writer ID
     */
    int wid = daq_recv_id(daq);
    if (wid < 0) {
	printf("Error reading writerID in daq_recv_block\n");
	return DAQD_ERROR;
    }

    /*------------------------------  Get the offline flag
     */
    int bl = daq_recv_block_num(daq);
    if (bl < 0) {
	printf("Error reading offline flag in daq_recv_block\n");
	rc = DAQD_ERROR;
    }
    return rc;
}


struct nds2_chan_status {
    int status;
    int offset;
    int data_type;
    float rate;
    float signal_offset;
    float signal_slope;
};

/*  nds2_receive_reconfigure(daq_t* daq, long block_len)
 *
 *  Receive a reconfigure block. receive_reconfigure is received after the 
 *  block header has been read in. The block length does not include the
 *  header length.
 */
int 
nds2_receive_reconfigure(daq_t* daq, long block_len) {
    long nchannels = block_len / sizeof(struct nds2_chan_status);
    if (nchannels * sizeof(struct nds2_chan_status) != block_len) {
	fprintf (stderr, 
		 "Channel reconfigure block length has bad length (%li)\n",
		 block_len);
    }
    if (nchannels != daq->num_chan_request) return -1;

    int fd = daq->datafd;
    chan_req_t* list = daq->chan_req_list;
    int i;
    for (i = 0; i < nchannels; i++) {
	float rate;
	uint4_type dtype;
	if (read_uint4(fd, (uint4_type*)&list[i].status))  return DAQD_ERROR;
	if (read_uint4(fd, &list[i].offset))  return DAQD_ERROR;
	if (read_uint4(fd, &dtype))           return DAQD_ERROR;
	if (list[i].type == cUnknown) list[i].type = dtype >> 16;
	list[i].data_type = (dtype & 0xffff);
	if (read_float(fd, &rate))           return DAQD_ERROR;
	list[i].rate = rate;
	if (read_float(fd, &list[i].s.signal_offset))  return DAQD_ERROR;
	if (read_float(fd, &list[i].s.signal_slope))   return DAQD_ERROR;
    }
    return -2; 
}

/*
 *  User authentication
 */
int
nds2_startup(void) {
    int rc = 0;
#ifdef HAVE_SASL
    static int once=0;
    if (!once) {
	rc = sasl_client_init(0);
	once = 1;
    }
#endif
    return rc;
}
